/*
  *  
  *	see tcpha_be.c for infoes about TCPHA
  *
  */
  
#ifndef _TCPHA_BE_H
#define _TCPHA_BE_H

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/list.h>                 /* for struct list_head */
#include <linux/spinlock.h>             /* for struct rwlock_t */
#include <linux/skbuff.h>               /* for struct sk_buff */
#include <linux/ip.h>                   /* for struct iphdr */
#include <asm/atomic.h>                 /* for struct atomic_t */
#include <linux/sysctl.h>
#include <linux/net.h>
#include <linux/inet.h>
#include <linux/vmalloc.h>
#include <linux/unistd.h>
#include <linux/wait.h>
#include <asm/unistd.h>
#include <linux/list.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_arp.h>
#include <linux/if_ether.h>
#include <linux/if_arp.h>
#include <linux/file.h>
#include <linux/in.h>
#include <linux/random.h>
#include <linux/smp_lock.h>
#include <linux/proc_fs.h>
#include <linux/netdevice.h>		/* for struct neighbour; */
#include <net/dst.h>			/* for struct dst_entry */
#include <net/route.h>			/* for ip_route_output */
#include <net/tcp.h>
#include <net/udp.h>
#include <linux/types.h>
#include <linux/errno.h>          
#include <net/icmp.h>             /* for icmp_send */
#include <net/ip.h>
#include <net/route.h>

#include "regex/regex.h"

/*
 *      TCPHA_BE State Values
 */
enum {
	TCPHA_BE_S_ESTABLISHED = 0,
	TCPHA_BE_S_WILLHANDOFF,
	TCPHA_BE_S_HANDOFFING,
	TCPHA_BE_S_HANDOFFED,
	TCPHA_BE_S_TIME_WAIT,
	TCPHA_BE_S_LAST
};

enum {
	TCPHA_BE_S_AVAIL =0,
	TCPHA_BE_S_NOTAVAIL,
	TCPHA_BE_S_BUSY
};

/* infoes used by TCP connection Handoff */
struct tcpha_conn_info {
	__u32 			magic;		/* Magic number */
	__u32			connmagic;	/* Connection magic number */
	__u32			daddr;		/* Foreign IPv4 addr			*/
	__u32 			saddr;		/* Sending source			*/
	__u32			rcv_saddr;	/* Bound local IPv4 addr		*/
	__u16			dport;		/* Destination port			*/
	__u16			sport;		/* Source port				*/
	struct tcp_opt	af_tcp;
};

#define TCPHACONNINFOSIZE sizeof(struct tcpha_conn_info)

struct tcpha_fe_info {
	__u32 			magic;
	__u32 			caddr;
	__u16			cport;
	__u32			daddr;
	__u16			dport;
	__u32			ndaddr;
};

struct tcpha_be_config {
	/* the server IP address and/or port */
	__u32 raddr;            /* real server address */
	__u32 uaddr;		    /* unused address in cluster */
	__u32 vaddr;		    /* shared vitual server address in cluster */
	__u16 port;

	int startservers;
	int maxSpareServers;
	int minSpareServers;

	int maxClients;

	__u32 faddr;
	__u16 fport;
	
	/* address/port to redirect */
	__u32 redirect_addr;
	__u16 redirect_port;
};

struct tcpha_be_pkt_pool {
	struct sk_buff_head receive_queue;
	wait_queue_head_t wait;
};

union tcpha_be_tphdr {
	unsigned char *raw;
	struct tcphdr *th;
	struct icmphdr *icmph;
	__u16 *portp;
};

struct tcpha_be_service {

	/* server configuration */
	struct tcpha_be_config conf;

	/* 	real server list and rule list,
	  *	no need locks since once they are  initilized,
	  *	won't change later 
	  */
	struct list_head be_list;
	struct list_head rule_list;
	
	/* server control */
	int start;
	int stop;

	/* run-time variables */
	struct socket *mainsock;
	atomic_t workercount;	/* worker counter */
	atomic_t running;	/* running flag */
};

typedef struct {
	struct list_head list;
	struct tcpha_be *be;
}be_t;

typedef struct {
	struct list_head list;
	regex_t rx;
	struct list_head be_list;
	spinlock_t __be_list_lock;
}rule_t;

struct tcpha_be_worker {
	struct tcpha_be_service *svc;	/* service it belongs to */
	int pid;		/* pid of worker */
	volatile __u16 status;	/* worker status */
};

struct timeout_table {
	atomic_t refcnt;
	int scale;
	int timeout[TCPHA_BE_S_LAST+1];
};

/* a connection entry in hash */
struct tcpha_be_client_conn {
	struct list_head list;	/* d-linked list head */
	rwlock_t lock;
	
	__u32 	caddr;		/* client address */
	__u16  	cport;		/* client port */
	__u32 	daddr;		/* server address */
	__u16    	dport;		/* server port */
	__u32 	ndaddr;		/* new server address */

	atomic_t		refcnt;		/* reference count */
	struct timer_list	timer;		/* Expiration timer */
	volatile unsigned long	timeout;	/* timeout */
	struct timeout_table *timeout_table;

	volatile __u16          flags;          /* status flags */
	volatile __u16          state;          /* state info */
};

/* a connection to FE */
struct tcpha_be_conn {
	struct socket *fsock;	/* socket connected to FE */

	#ifdef CONFIG_TCPHA_BE_DEBUG
	int magic;
	#endif
};

/* cluster BE entry */
struct tcpha_be {
	struct list_head list;
	
	volatile __u16          state;          /* state info */
	__u32 addr;		/* IP address of real server */
	__u16 port;		/* port number of the service */
};

#define TCPHAMAGIC 		0x12968b9

/*
 *      TCPHA Connection Flags
 */
#define TCPHA_BE_CONN_F_FWD_MASK         0x0007    /* mask for the fwd methods */
#define TCPHA_BE_CONN_F_TUNNEL	      0x0002    /* tunneling */
#define TCPHA_BE_CONN_F_HASHED	      0x0040	/* hashed entry */

/*
 *     TCPHA_BE connection entry hash table
 */
#ifndef CONFIG_TCPHA_BE_TAB_BITS
#define CONFIG_TCPHA_BE_TAB_BITS   12
#endif
/* make sure that TCPHA_BE_CONN_TAB_BITS is located in [8, 20] */
#if CONFIG_TCPHA_BE_TAB_BITS < 8
#define TCPHA_BE_CONN_TAB_BITS	8
#endif
#if CONFIG_TCPHA_BE_TAB_BITS > 20
#define TCPHA_BE_CONN_TAB_BITS	20
#endif
#if 8 <= CONFIG_TCPHA_BE_TAB_BITS && CONFIG_TCPHA_BE_TAB_BITS <= 20
#define TCPHA_BE_CONN_TAB_BITS	CONFIG_TCPHA_BE_TAB_BITS
#endif
#define TCPHA_BE_CONN_TAB_SIZE     (1 << TCPHA_BE_CONN_TAB_BITS)
#define TCPHA_BE_CONN_TAB_MASK     (TCPHA_BE_CONN_TAB_SIZE - 1)

#define INFO(msg...) printk(KERN_INFO "TCPHA_BE: " msg)
#ifdef CONFIG_TCPHA_BE_DEBUG
#define DBG(msg...) printk(KERN_INFO "TCPHA_BE: " msg)
#define DBG_RL(msg...) 					\
    do {						\
	    if (net_ratelimit())			\
		    printk(KERN_INFO "TCPHA_BE: " msg);	\
    } while (0)
#else
#define DBG(msg...) do {} while (0)
#define DBG_RL(msg...) do {} while (0)
#endif
#define ERR(msg...) printk(KERN_ERR "TCPHA_BE: " msg)
#define ERR_RL(msg...)				\
    do {						\
	    if (net_ratelimit())			\
		    printk(KERN_ERR "TCPHA_BE: " msg);	\
    } while (0)

/*
 *  Connection hash table: for input and output packets lookups of TCPHA_BE
 */
struct list_head *tcpha_be_conn_tab;

/* SLAB cache for TCPHA_BE connections */
kmem_cache_t *tcpha_be_client_conn_cachep;

/* counter for current TCPHA_BE connections */
atomic_t tcpha_be_conn_count;

/* random value for TCPHA_BE connection hash */
unsigned int tcpha_be_conn_rnd;

/* SLAB cache for TCPHA_BE connections */
kmem_cache_t *tcpha_be_conn_cachep;
		
struct tcpha_be_service tbs;

struct tcpha_be_pkt_pool pkt_buf;

#endif
