/*
  *  
  *	see tcpha_be.c for infoes about TCPHA
  *
  */
  
#include "tcpha_be.h"
#include "tcpha_be_hash.h"
#include "tcpha_be_con.h"

static unsigned int noarp_infilter (unsigned int hook,
                                    struct sk_buff **pskb,
                                    const struct net_device *indev,
                                    const struct net_device *outdev,
                                    int (*okfn)(struct sk_buff *))
   {
   struct sk_buff * skb   = *pskb;
   struct ethhdr  * peth  = skb->mac.ethernet;
   struct arphdr  * parp  = skb->nh.arph;
   __u8 * pch   = ((__u8 *)skb->nh.arph) + sizeof (struct arphdr);
   __u32 src,dst;
   
   int apiRet = NF_ACCEPT;


   if ((peth == NULL) ||
       (peth->h_proto != __constant_htons (ETH_P_ARP)) ||
       (parp->ar_hln  != (unsigned char) 6) ||
       (parp->ar_pln  != (unsigned char) 4) ||
       (parp->ar_op   != __constant_htons (ARPOP_REQUEST)))
      {
          return apiRet;
      }
   
   src=*((uint32_t*)(pch + parp->ar_hln));
   dst=*((uint32_t*)(pch + parp->ar_hln + parp->ar_pln + parp->ar_hln));

   if (dst==tbs.conf.vaddr)
   	apiRet=NF_DROP;

   if (src==tbs.conf.vaddr)
   	*((uint32_t*)(pch + parp->ar_hln))=tbs.conf.uaddr;

   return apiRet;
   }

/****************************************************************************/
/* Output ARP Filtering
*****************************************************************************/
static unsigned int noarp_outfilter (unsigned int hook,
                                     struct sk_buff **pskb,
                                     const struct net_device *indev,
                                     const struct net_device *outdev,
                                     int (*okfn)(struct sk_buff *))
   {
   struct sk_buff * skb   = *pskb;
   struct arphdr  * parp  = skb->nh.arph;
   __u8 * pch   = ((__u8 *)skb->nh.arph) + sizeof (struct arphdr);
   __u32 src,dst;
   int apiRet = NF_ACCEPT;

   if ((parp->ar_hln  != (unsigned char) 6) ||
       (parp->ar_pln  != (unsigned char) 4))
      {
      return apiRet;
      }
   src=*((uint32_t*)(pch + parp->ar_hln));
   dst=*((uint32_t*)(pch + parp->ar_hln + parp->ar_pln + parp->ar_hln));

   if (dst==tbs.conf.uaddr)
   	{
          *((uint32_t*)(pch + parp->ar_hln + parp->ar_pln + parp->ar_hln))=tbs.conf.vaddr;
          return apiRet;
   	}

   if (parp->ar_op == __constant_htons (ARPOP_REPLY))
   {
        if (src==tbs.conf.vaddr)
        	apiRet=NF_DROP;
   }

   else if (parp->ar_op == __constant_htons (ARPOP_REQUEST))
      {
           if (src==tbs.conf.vaddr)
           	*((uint32_t*)(pch + parp->ar_hln)) = tbs.conf.raddr;
      }
   return apiRet;
   }

/* get connection state,only called from bh */
static inline int get_conn_state(struct tcpha_be_client_conn *cp)
{
	int ret;
	
	read_lock(&cp->lock);
	ret=cp->state;
	read_unlock(&cp->lock);

	return ret;
}

/*
 *	transport layer header checking
 */
static inline int tcpha_be_header_check(struct sk_buff *skb, int proto, int ihl)
{
	int len;

	switch (proto) {
	case IPPROTO_TCP:
		len = ihl + sizeof(struct tcphdr);
		/* we don't care about TCP options */
		break;
	default:
		len = 0;
	}

	/* guarantee protocol header available in skb data area */
	if (!pskb_may_pull(skb, len))
		return -1;
	else
		return 0;
}

static inline void tell_advwkr_dataready(void)
{
	if ( waitqueue_active(&pkt_buf.wait))
		wake_up_interruptible(&pkt_buf.wait);
}

static inline struct rtable *tcpha_be_get_out_rt(__u32 dst, u32 rtos)
{
	struct rtable *rt;			/* Route to the other host */

       if (ip_route_output(&rt, dst, 0, rtos, 0)) {
		return NULL;
	}

	return rt;
}

static int tcpha_be_tunnel_xmit(struct sk_buff *skb, struct tcpha_be_client_conn *cp)
{
	struct rtable *rt;			/* Route to the other host */
	struct net_device *tdev;		/* Device to other host */
	struct iphdr  *old_iph = skb->nh.iph;
	u8     tos = old_iph->tos;
	u16    df = old_iph->frag_off;
	struct iphdr  *iph;			/* Our new IP header */
	int    max_headroom;			/* The extra header space needed */
	int    mtu;

	if (skb->protocol != __constant_htons(ETH_P_IP)) {
		DBG_RL("tcpha_be_tunnel_xmit(): protocol error, "
			     "ETH_P_IP: %d, skb protocol: %d\n",
			     __constant_htons(ETH_P_IP), skb->protocol);
		goto tx_error;
	}

	if (!(rt = tcpha_be_get_out_rt(cp->ndaddr, RT_TOS(tos))))
		goto tx_error_icmp;

	tdev = rt->u.dst.dev;

	mtu = rt->u.dst.pmtu - sizeof(struct iphdr);
	if (mtu < 68) {
		ip_rt_put(rt);
		DBG_RL("tcpha_be_tunnel_xmit(): mtu less than 68\n");
		goto tx_error;
	}
	if (skb->dst && mtu < skb->dst->pmtu)
		skb->dst->pmtu = mtu;

	df |= (old_iph->frag_off&__constant_htons(IP_DF));

	if ((old_iph->frag_off&__constant_htons(IP_DF))
	    && mtu < ntohs(old_iph->tot_len)) {
		icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
		ip_rt_put(rt);
		DBG_RL("tcpha_be_tunnel_xmit(): frag needed\n");
		goto tx_error;
	}

	/* update checksum because skb might be defragmented */
	ip_send_check(old_iph);

	skb->h.raw = skb->nh.raw;

	/*
	 * Okay, now see if we can stuff it in the buffer as-is.
	 */
	max_headroom = (((tdev->hard_header_len+15)&~15)+sizeof(struct iphdr));

	if (skb_headroom(skb) < max_headroom
	    || skb_cloned(skb) || skb_shared(skb)) {
		struct sk_buff *new_skb =
			skb_realloc_headroom(skb, max_headroom);
		if (!new_skb) {
			ip_rt_put(rt);
			kfree_skb(skb);
			ERR_RL("tcpha_be_tunnel_xmit(): no memory\n");
			return -EINVAL;
		}
		kfree_skb(skb);
		skb = new_skb;
		old_iph = skb->nh.iph;
	}

	skb->nh.raw = skb_push(skb, sizeof(struct iphdr));
	memset(&(IPCB(skb)->opt), 0, sizeof(IPCB(skb)->opt));

	/* drop old route */
	dst_release(skb->dst);
	skb->dst = &rt->u.dst;

	/*
	 *	Push down and install the IPIP header.
	 */
	iph			=	skb->nh.iph;
	iph->version		=	4;
	iph->ihl		=	sizeof(struct iphdr)>>2;
	iph->frag_off		=	df;
	iph->protocol		=	IPPROTO_IPIP;
	iph->tos		=	tos;
	iph->daddr		=	rt->rt_dst;
	iph->saddr		=	tbs.conf.raddr;
	iph->ttl		=	old_iph->ttl;
	iph->tot_len		=	htons(skb->len);
	ip_select_ident(iph, &rt->u.dst, NULL);
	ip_send_check(iph);

	skb->ip_summed = CHECKSUM_NONE;
#ifdef CONFIG_NETFILTER_DEBUG
	skb->nf_debug = 1 << NF_IP_LOCAL_OUT;
#endif /* CONFIG_NETFILTER_DEBUG */
	ip_send(skb);

	return NF_STOLEN;

  tx_error_icmp:
	dst_link_failure(skb);
  tx_error:
	kfree_skb(skb);
	return NF_STOLEN;
}

static inline int checksum_check(struct sk_buff *skb,struct iphdr *iph,struct tcphdr *th)
{
	int len=skb->len-(iph->ihl << 2);
	
	switch (skb->ip_summed) {
	case CHECKSUM_NONE:
		skb->csum = csum_partial((void *)th, len, 0);
	case CHECKSUM_HW:
		if (csum_tcpudp_magic(iph->saddr, iph->daddr,
				      len, IPPROTO_TCP, skb->csum)) {
			return NF_DROP;
		}
		break;
	default:/* CHECKSUM_UNNECESSARY */
		break;
	}
	return NF_ACCEPT;
}

struct tcpha_be *
tcpha_be_http_schedule(char *buf,int len,struct tcpha_be_service *);

static unsigned int tcpha_be_in(unsigned int hooknum,
			     struct sk_buff **skb_p,
			     const struct net_device *in,
			     const struct net_device *out,
			     int (*okfn)(struct sk_buff *))
{
	struct sk_buff	*skb = *skb_p;
	struct iphdr	*iph = skb->nh.iph;
	union tcpha_be_tphdr h;
	struct tcpha_be_client_conn *cp;
	int ihl;

	/*
	 *	Big tappo: only PACKET_HOST (nor loopback neither mcasts)
	 *	... don't know why 1st test DOES NOT include 2nd (?)
	 */
	if (skb->pkt_type != PACKET_HOST || skb->dev == &loopback_dev) {
		return NF_ACCEPT;
	}

	/* let it go if other IP protocols */
	if (iph->protocol != IPPROTO_TCP)
		return NF_ACCEPT;

	/* make sure that protocol header available in skb data area,
	   note that skb data area may be reallocated. */
	ihl = iph->ihl << 2;
	if (tcpha_be_header_check(skb, iph->protocol, ihl) == -1)
		return NF_DROP;
	iph = skb->nh.iph;
	h.raw = skb->h.raw=(char*) iph + ihl;
	
	/*
	 * Check if the packet belongs to an existing connection entry
	 */
	cp = tcpha_be_conn_in_get(iph->saddr, h.portp[0],
			       iph->daddr, h.portp[1]);

	if (!cp)
		return NF_ACCEPT;

	int ret=NF_ACCEPT,state;
	int thl=h.th->doff<<2;
	
	state=get_conn_state(cp);
	switch (state) {
		case TCPHA_BE_S_ESTABLISHED:
			if (h.th->fin||h.th->rst) {
				tcpha_be_set_state_timeout(cp, TCPHA_BE_S_TIME_WAIT);
				break;
			}
			if (skb->len<=ihl+thl) {
				DBG("no content,skipped\n");
				break;
			}
			ret=checksum_check(skb,iph,h.th);
			if (ret!=NF_ACCEPT) {
	     			DBG("checksum error\n");
				break;
			}
			if (!pskb_may_pull(skb,skb->len)) {
				DBG("skb error\n");
				break;
			}
			struct tcpha_be *be=tcpha_be_http_schedule(skb->data+ihl+thl,skb->len-ihl-thl,&tbs);
			if (!be) {
				DBG("no BE available\n");
				break;
			}
			if (be->addr==tbs.conf.raddr) {
				DBG("the same BE\n");
				break;
			}
			skb_unlink(skb);
			__skb_pull(skb,ihl+thl);
			/* ugly:) */
			h.th->seq=(__u32)be;
			h.th->ack_seq=(__u32)cp;
			__skb_queue_tail(&pkt_buf.receive_queue, skb);
			tell_advwkr_dataready();
			tcpha_be_set_state_timeout(cp,TCPHA_BE_S_WILLHANDOFF);
			ret=NF_STOLEN;
			break;
		case TCPHA_BE_S_WILLHANDOFF:
			if (h.th->fin||h.th->rst) {
				tcpha_be_set_state_timeout(cp, TCPHA_BE_S_TIME_WAIT);
				break;
			}
			if (skb->len>ihl+thl) {
				DBG("droped new http request\n");
				ret=NF_DROP;
			}
			break;
		case TCPHA_BE_S_HANDOFFING:
			ret=NF_DROP;
			break;
		case TCPHA_BE_S_HANDOFFED:
			skb_unlink(skb);
			ret=tcpha_be_tunnel_xmit(skb,cp);
			break;
		default:
			break;
	}
	tcpha_be_conn_put(cp);
	return ret;
}

/* make sure that */
static unsigned int tcpha_be_out(unsigned int hooknum,
			      struct sk_buff **skb_p,
			      const struct net_device *in,
			      const struct net_device *out,
			      int (*okfn)(struct sk_buff *))
{
	struct sk_buff  *skb = *skb_p;
	struct iphdr	*iph;
	union tcpha_be_tphdr h;
	struct tcpha_be_client_conn *cp;
	int ihl;

	iph = skb->nh.iph;

	/* let it go if other IP protocols */
	if (iph->protocol != IPPROTO_TCP)
		return NF_ACCEPT;

	/* make sure that protocol header available in skb data area,
	   note that skb data area may be reallocated. */
	ihl = iph->ihl << 2;
	if (tcpha_be_header_check(skb, iph->protocol, ihl) == -1)
		return NF_DROP;

	iph = skb->nh.iph;
	h.th = (struct tcphdr *) iph + ihl;

	/*
	 *	Check if the packet belongs to an old entry
	 */
	cp = tcpha_be_conn_out_get(iph->saddr, h.portp[0],
				iph->daddr, h.portp[1]);
	if (!cp) {
		return NF_ACCEPT;
	}

	int ret=NF_ACCEPT,state;

	state=get_conn_state(cp);
	switch (state) {
		case TCPHA_BE_S_ESTABLISHED:
		case TCPHA_BE_S_WILLHANDOFF:
			if (h.th->fin||h.th->rst) {
				tcpha_be_set_state_timeout(cp, TCPHA_BE_S_TIME_WAIT);
				break;
			}
			break;
		case TCPHA_BE_S_TIME_WAIT:
			break;
		default:
			ret=NF_STOLEN;
			break;
	}
	tcpha_be_conn_put(cp);
	return ret;
}

/****************************************************************************/
/* Hooks Table
*****************************************************************************/
struct nf_hook_ops noarp_nf_ops_in =
   {{ NULL, NULL }, noarp_infilter, NF_ARP, NF_ARP_IN, 0};
struct nf_hook_ops noarp_nf_ops_out =
   {{ NULL, NULL }, noarp_outfilter, NF_ARP, NF_ARP_OUT, 0};
struct nf_hook_ops tcpha_be_in_ops = {
	{ NULL, NULL },
	tcpha_be_in, PF_INET, NF_IP_LOCAL_IN, 100
};
struct nf_hook_ops tcpha_be_out_ops = {
	{ NULL, NULL },
	tcpha_be_out, PF_INET, NF_IP_FORWARD, 100
};
