/*
  *  
  *	see tcpha_be.c for infoes about TCPHA
  *
  */

#include "tcpha_be.h"
#include "tcpha_be_file.h"

#define DEFBEPORT					666
#define DEFFEPORT					666
#define DEFSTARTSERVERS			5
#define DEFMAXSPARESERVERS		5
#define DEFMINSPARESERVERS			2
#define DEFMAXCLIENTS				256
#define DEFREDIRECTPORT			80

ssize_t
ke_fgets(struct file *file, char *buf, size_t n)
{
#define EOF (-1)

	ssize_t count;
	mm_segment_t oldfs;
	struct inode *inode;

	if (!file || IS_ERR(file) || !buf || n <= 0) {
		goto out_error;
	}
	if (!file || !file->f_dentry || !file->f_dentry->d_inode)
		goto out_error;

	inode = file->f_dentry->d_inode;

	if (!(file->f_mode & FMODE_READ))
		goto out_error;

	if (file->f_op && file->f_op->read) {
		oldfs = get_fs();
		set_fs(KERNEL_DS);
		count = 0;

		if (file->f_op->read(file, buf, 1, &file->f_pos) == 0)
			goto out;

		if (*buf == EOF)
			goto out;
		count = 1;
		while (*buf != '\n' && *buf != '\0' && *buf != EOF
		       && count < n && *buf != '\r'
		       && file->f_pos <= inode->i_size) {
			buf += 1;
			count += 1;
			if (file->f_op->read(file, buf, 1, &file->f_pos) <=
			    0) {
				count -= 1;
				break;
			}
		}
	} else
		goto out_error;
	if (*buf != '\0') {
		buf += 1;
		*buf = '\0';
	} else
		count -= 1;
      out:
	set_fs(oldfs);
	return count;
      out_error:
	return -1;
}

#define MAX_STRING_LEN 256

int
get_next_token(struct config_file *cf, int *newline)
{
	char *p, *tp;
	char c;
	int tem;

	*newline = 0;
	for (p = cf->position;; p++) {
		if (p >= cf->lineEnd) {
			if ((tem =
			     ke_fgets(cf->fd, cf->line,
				      MAX_STRING_LEN)) < 0) {
				return -1;
			}

			*newline = 1;
			if (tem == 0)
				return 1;
			p = cf->line;
			cf->lineEnd = cf->line + strlen(cf->line);
			cf->lineNum++;
		}

		c = *p;
		if (c == ' ')
			continue;

		if (c == '#') {
			p = cf->lineEnd;
			continue;
		}

		if (c == '\r' || c == '\n')
			continue;

		/* get token here */
		tp = cf->token;
		if (c == '"') {
			/* get a quoted string here */
			for (p++; *p != '"';) {
				c = *tp++ = *p++;
				if (c == '\\' && *p == '"')
					*tp++ = *p++;
				if (p >= cf->lineEnd)
					return -1;
			}
			p++;	/* skip the last quote symbol */
		} else {
			do {
				*tp++ = *p++;
			} while (*p != ' ' && *p != '\r' && *p != '\n'
				 && p < cf->lineEnd);
		}

		*tp = '\0';
		cf->position = p;
		return 0;
	}
}

struct config_file *
open_configfile(const char *filename)
{
	struct config_file *cf = NULL;
	struct file *fd;

	if (!
	    (cf =
	     (struct config_file *) kmalloc(sizeof(*cf), GFP_KERNEL))) {
		ERR("Alloc memory error\n");
		goto out;
	}

	cf->name = strdup(filename);
	cf->line = (char *) kmalloc(MAX_STRING_LEN, GFP_KERNEL);
	if (!cf->line) {
		ERR("Alloc memory error\n");
		goto out_error;
	}
	cf->token = (char *) kmalloc(MAX_STRING_LEN, GFP_KERNEL);
	if (!cf->token) {
		ERR("Alloc memory error\n");
		goto out_error;
	}
	fd = filp_open(filename, O_RDONLY, 0);
	if (!fd || IS_ERR(fd))
		goto out_error;
	cf->fd = fd;
	cf->position = cf->line;
	cf->lineEnd = cf->line;
	cf->lineNum = 1;
      out:
	return cf;
      out_error:
	if (cf->line) {
		kfree(cf->line);
		cf->line = NULL;
	}
	kfree(cf);
	cf = NULL;
	goto out;
}

void
close_configfile(struct config_file *cf)
{
	filp_close(cf->fd, current->files);
	if (cf->line)
		kfree(cf->line);
	if (cf->name)
		kfree(cf->name);
	if (cf->token)
		kfree(cf->token);
	kfree(cf);
	return;
}

int
str2int(char *p, int len, __u32 * num)
{
	int i;
	*num = 0;
	if (len <= 0)
		goto out_error;
	for (i = 0; i < len; i++) {
		if (*(p + i) < '0' || *(p + i) > '9')
			goto out_error;
		(*num) *= 10;
		(*num) += *(p + i) - '0';
	}
	return 0;
      out_error:
	return -1;
}

int init_be_list(struct tcpha_be_service *svc,__u32 addr,__u16 port)
{
	struct tcpha_be *be=kmalloc(sizeof(*be),GFP_KERNEL);
	
	if (!be) {
		ERR("no memory\n");
		return -ENOMEM;
	}
	
	/* init a BE */
	be->addr=addr;
	be->port=port;
	be->state=TCPHA_BE_S_AVAIL;
	list_add(&be->list,&svc->be_list);
	
	return 0;
}

static inline rule_t *new_rule(void)
{
	rule_t *rule=(rule_t *)kmalloc(sizeof(*rule),GFP_KERNEL);
	if (!rule)
		return NULL;
	INIT_LIST_HEAD(&rule->list);
	INIT_LIST_HEAD(&rule->be_list);
	spin_lock_init(&rule->__be_list_lock);
	return rule;
}

static inline struct tcpha_be *lookup_be(struct tcpha_be_service *svc,__u32 addr)
{
	struct list_head *l;
	struct tcpha_be *be=NULL;

	list_for_each(l, &svc->be_list) {
		be = list_entry(l, struct tcpha_be, list);
		if (be->addr==addr)
			break;
		be=NULL;
	}
	return be;
}

int add_rule(struct tcpha_be_service *svc,char *token,__u32 addr)
{
	struct list_head *l;
	rule_t *rule=NULL;
	struct tcpha_be *tmp;
	be_t *be;
	
	if (!(tmp=lookup_be(svc,addr))) {
		ERR("the real server:%u.%u.%u.%u used by rule:%s not defined\n",NIPQUAD(addr),token);
		goto out_error;
	}
	be=(be_t *)kmalloc(sizeof(*be),GFP_KERNEL);
	if (!be)
		goto out_error;
	list_for_each(l, &svc->rule_list) {
		rule = list_entry(l, rule_t, list);
		if (!regexec(&rule->rx,token, 0, NULL, 0)) {
			break;
		}
		rule=NULL;
	}
	if (!rule) {
		if (!(rule=new_rule()))
			goto out_error;
		if (regcomp(&rule->rx,token, REG_EXTENDED | REG_NOSUB)) {
		        ERR("pattern compiling error\n");
			 kfree(rule);
		        goto out_error;
	       }
		list_add(&rule->list,&svc->rule_list);
	}
	INIT_LIST_HEAD(&be->list);
	be->be=tmp;
	list_add(&be->list,&rule->be_list);
	return 0;
	out_error:
	return -1;
}

int
handle_file(struct config_file *cf, struct tcpha_be_service *svc)
{
	int newline = 0, i;
	struct tcpha_be_config *conf=&svc->conf;
	__u32 addr,port;
	char *commands[13] = { "raddr", "uaddr","vaddr","port","startservers","maxspareservers",
			"minspareservers","maxclients","redirectaddr","redirectport","fe","server","rule"};

	/* step 1:get "tcpha_be {" */
	if ((i = get_next_token(cf, &newline)) == -1 || newline == 0) {
		goto out_error;
	}
	if (i == 1)
		goto out_error;
	if (strcasecmp("tcpha_be",cf->token))
		goto out_error;

	if (get_next_token(cf, &newline) || newline == 1) {
		goto out_error;
	}
	if (strcmp("{", cf->token))
		goto out_error;

	/*step 2:handle commands */
      handle_command:
	if ((i = get_next_token(cf, &newline)) == -1 || newline == 0) {
		goto out_error;
	}
	if (i == 1)
		goto out_error;
	for (i = 0; i < 13; i++) {
		if (!strcasecmp(commands[i],cf->token)) {
			if (get_next_token(cf, &newline) || newline == 1)
				goto out_error;
			if (strcmp("=", cf->token))
				goto out_error;
			if (get_next_token(cf, &newline) || newline == 1)
				goto out_error;
			switch (i) {
			case 0:
				if (isavalidip(cf->token,&conf->raddr)) {
					ERR("raddr specified is invalid\n");
					goto out_error;
				}
				break;
			case 1:
				if (isavalidip(cf->token,&conf->uaddr)) {
					ERR("uaddr specified is invalid\n");
					goto out_error;
				}
				break;
			case 2:
				if (isavalidip(cf->token,&conf->vaddr)) {
					ERR("vaddr specified is invalid\n");
					goto out_error;
				}
				break;
			case 3:
				if (str2int
				    (cf->token, strlen(cf->token), &port)||port>65534) {
				    	ERR("port specified is invalid\n");
					goto out_error;
				}
				conf->port = htons(port);
				break;
			case 4:
				if (str2int
				    (cf->token, strlen(cf->token), &conf->startservers)) {
				    	ERR("startservers specified is invalid\n");
					goto out_error;
				}
				if (conf->startservers>TCPHA_BE_MAX_DAEMONS) {
					ERR("server num specified in config file is invalid,max num is:%d,ignore it.\n",
					     conf->startservers=TCPHA_BE_MAX_DAEMONS);
				}
				break;
			case 5:
				if (str2int(cf->token, strlen(cf->token), &conf->maxSpareServers)) {
					ERR("maxspareservers specified is invalid\n");
					goto out_error;
				}
				if (conf->maxSpareServers>TCPHA_BE_MAX_DAEMONS) {
					ERR("max spare server num specified in config file is invalid,max num is:%d,ignore it.\n",
					     conf->maxSpareServers=TCPHA_BE_MAX_DAEMONS);
				}
				break;
			case 6:
				if (str2int(cf->token, strlen(cf->token), &conf->minSpareServers)) {
					ERR("minspareservers specified is invalid\n");
					goto out_error;
				}
				if (conf->minSpareServers<0) {
					ERR("Min spare server num specified in config file is invalid,min num is:%d,ignore it.\n",
					     conf->minSpareServers=0);
				}
				break;
			case 7:
				if (str2int(cf->token, strlen(cf->token), &conf->maxClients)) {
					ERR("maxclients specified is invalid\n");
					goto out_error;
				}
				break;
			case 8:
				if (isavalidip(cf->token,&conf->redirect_addr)) {
					ERR("redirectaddr specified is invalid\n");
					goto out_error;
				}
				break;
			case 9:
				if (str2int(cf->token, strlen(cf->token), &port)||port>65534) {
					ERR("redirectport specifed is invalid\n");
					goto out_error;
				}
				conf->redirect_port = htons(port);
				break;
			case 10:
				if (isavalidip(cf->token,&addr)) {
					ERR("FE address specified is invalid\n");
					goto out_error;
				}
				if (get_next_token(cf, &newline) || newline == 1) {
					ERR("FE port not specified\n");
					goto out_error;
				}
				if (str2int
				    (cf->token, strlen(cf->token), &port)||port>65534) {
				    	ERR("FE port specified is invalid\n");
					goto out_error;
				}
				conf->faddr=addr;
				conf->fport=htons(port);
				break;
			case 11:
				if (isavalidip(cf->token,&addr)) {
					ERR("server address specified is invalid\n");
					goto out_error;
				}
				if (get_next_token(cf, &newline) || newline == 1) {
					ERR("server port not specified\n");
					goto out_error;
				}
				if (str2int
				    (cf->token, strlen(cf->token), &port)||port>65534) {
				    	ERR("server port specified is invalid\n");
					goto out_error;
				}
				if (init_be_list(svc,addr,htons(port)))
					goto out_error;
				break;
			 case 12:
           	    	    	if (isavalidip(cf->token,&addr)) {
					 ERR("server address specified is invalid\n");
           	    	    		 goto out_error;
           	    	    	}
		              if (get_next_token(cf,&newline)||newline==1)
		                       goto out_error;
					   
		              if (add_rule(svc,cf->token,addr))
					  goto out_error;
			   	break;
			default:
				break;
			}
			break;
		}
	}
	if (i >=13)
		goto out;
	goto handle_command;

      out:
	/* step 3:read the "}",in fact leave it be */
	return 0;
      out_error:
	ERR("wrong in parsing config file\n");
	return -1;
}

int config_svc_use_def(struct tcpha_be_service *svc)
{
	struct tcpha_be_config *conf=&svc->conf;

	if (!conf->raddr) {
		ERR("BE real server address not specified\n");
		goto out_error;
	}
	if (!conf->uaddr) {
		ERR("BE unused server address not specified\n");
		goto out_error;
	}
	if (!conf->vaddr) {
		ERR("BE virtual server address not specified\n");
		goto out_error;
	}
	if (!conf->faddr) {
		ERR("FE addr not specified\n");
		goto out_error;
	}
	if (!conf->port)
		conf->port=htons(DEFBEPORT);
	if (!conf->fport)
		conf->fport=htons(DEFFEPORT);
	if (!conf->startservers) {
		conf->startservers=conf->maxSpareServers?conf->maxSpareServers:DEFSTARTSERVERS;
	}
	if (!conf->minSpareServers) {
		conf->minSpareServers=DEFMINSPARESERVERS;
	}
	if (!conf->maxSpareServers)
		conf->maxSpareServers=conf->startservers>DEFMAXSPARESERVERS?DEFMAXSPARESERVERS:conf->startservers;
	if (!conf->maxClients)
		conf->maxClients=DEFMAXCLIENTS;
	if (!conf->redirect_addr)
		conf->redirect_addr=conf->raddr;
	if (!conf->redirect_port)
		conf->redirect_port=htons(DEFREDIRECTPORT);
	return 0;
	out_error:
	return -1;
}

#ifdef CONFIG_TCPHA_BE_DEBUG
void show_svc_config(struct tcpha_be_service *svc)
{
	struct tcpha_be_config *conf=&svc->conf;
	struct list_head *l;
	struct tcpha_be *be;
	
	INFO("BE config:\n"
		"raddr:%u.%u.%u.%u\n"
		"uaddr:%u.%u.%u.%u\n"
		"vaddr:%u.%u.%u.%u\n"
		"port:%d\n"
		"startservers:%d\n"
		"maxSpareServers:%d\n"
		"minSpareServers:%d\n"
		"maxClients:%d\n"
		"FE address:%u.%u.%u.%u:%d\n"
		"redirect_addr:%u.%u.%u.%u\n"
		"redirect_port:%d\n",
		NIPQUAD(conf->raddr),NIPQUAD(conf->uaddr),NIPQUAD(conf->vaddr),ntohs(conf->port),
		conf->startservers,conf->maxSpareServers,conf->minSpareServers,
		conf->maxClients,NIPQUAD(conf->faddr),ntohs(conf->fport),NIPQUAD(conf->redirect_addr),ntohs(conf->redirect_port));
	INFO("real servers:\n");
	list_for_each(l, &svc->be_list) {
		be = list_entry(l, struct tcpha_be, list);
		INFO("%u.%u.%u.%u:%d\n",NIPQUAD(be->addr),ntohs(be->port));
	}
}
#else
#define show_svc_config(x) do {} while (0)
#endif

int
config_tcpha_be(struct  tcpha_be_service *svc)
{
	struct config_file *cf;
	int ret=0;
	
	if (!(cf = open_configfile(DEFCONFIGFILE))) {
		ERR("open config file:%s error\n",DEFCONFIGFILE);
		ret=-1;
		goto out;
	}
	INIT_LIST_HEAD(&svc->be_list);
	INIT_LIST_HEAD(&svc->rule_list);
	ret=handle_file(cf, svc);
	close_configfile(cf);
	if (ret)
		goto out;
	ret=config_svc_use_def(svc);
	if (!ret)
		show_svc_config(svc);
	tcpha_be_conn_cachep = kmem_cache_create("tcpha_be_conn",
					      sizeof(struct tcpha_be_conn), 0,
					      SLAB_HWCACHE_ALIGN, NULL, NULL);
	if (!tcpha_be_conn_cachep) {
		ret=-ENOMEM;
	}
      out:
	return ret;
}
