/*
  *  
  *	see tcpha_be.c for infoes about TCPHA
  *
  */
  
#include "tcpha_be.h"

struct sock *fake_sock(struct sock *sk,struct tcpha_conn_info *msg)
{
	struct sock *newsk = sk_alloc(PF_INET, GFP_KERNEL, 0);

	if(newsk != NULL) {
		struct tcp_opt *newtp,*oldtp=&(msg->af_tcp);
#ifdef CONFIG_FILTER
		struct sk_filter *filter;
#endif

		memcpy(newsk, sk, sizeof(*newsk));
		newsk->state = TCP_ESTABLISHED;

		/* SANITY */
		newsk->pprev = NULL;
		newsk->prev = NULL;

		/* Clone the TCP header template */
		newsk->dport = msg->dport;
		newsk->sport=msg->sport;
		sock_lock_init(newsk);
		bh_lock_sock(newsk);

		newsk->dst_lock	= RW_LOCK_UNLOCKED;
		atomic_set(&newsk->rmem_alloc, 0);
		skb_queue_head_init(&newsk->receive_queue);
		atomic_set(&newsk->wmem_alloc, 0);
		skb_queue_head_init(&newsk->write_queue);
		atomic_set(&newsk->omem_alloc, 0);
		newsk->wmem_queued = 0;
		newsk->forward_alloc = 0;

		newsk->done = 0;
		newsk->userlocks = sk->userlocks & ~SOCK_BINDPORT_LOCK;
		newsk->proc = 0;
		newsk->backlog.head = newsk->backlog.tail = NULL;
		newsk->callback_lock = RW_LOCK_UNLOCKED;
		skb_queue_head_init(&newsk->error_queue);
		newsk->write_space = tcp_write_space;
#ifdef CONFIG_FILTER
		if ((filter = newsk->filter) != NULL)
			sk_filter_charge(newsk, filter);
#endif

		/* Now setup tcp_opt */
		newtp = &(newsk->tp_pinfo.af_tcp);
		*newtp=*oldtp;

		tcp_prequeue_init(newtp);


		/* So many TCP implementations out there (incorrectly) count the
		 * initial SYN frame in their delayed-ACK and congestion control
		 * algorithms that we must have the following bandaid to talk
		 * efficiently to them.  -DaveM
		 */

		newtp->ca_state = TCP_CA_Open;
		tcp_init_xmit_timers(newsk);
		skb_queue_head_init(&newtp->out_of_order_queue);
		newtp->send_head = NULL;

		newtp->listen_opt = NULL;
		newtp->accept_queue = newtp->accept_queue_tail = NULL;
		/* Deinitialize syn_wait_lock to trap illegal accesses. */
		memset(&newtp->syn_wait_lock, 0, sizeof(newtp->syn_wait_lock));

		/* Back to base struct sock members. */
		newsk->err = 0;
		newsk->priority = 0;
		atomic_set(&newsk->refcnt, 2);
#ifdef INET_REFCNT_DEBUG
		atomic_inc(&inet_sock_nr);
#endif
		atomic_inc(&tcp_sockets_allocated);

		if (newsk->keepopen)
			tcp_reset_keepalive_timer(newsk, TCP_KEEPALIVE_TIME);
		newsk->socket = NULL;
		newsk->sleep = NULL;

		TCP_INC_STATS_BH(TcpPassiveOpens);
	}
	return newsk;
}

static struct dst_entry* get_route(struct sock *sk, struct tcpha_conn_info *msg)
{
	struct rtable *rt;

	if(ip_route_output(&rt, msg->daddr,msg->saddr,
			   RT_CONN_FLAGS(sk), sk->bound_dev_if)) {
		IP_INC_STATS_BH(IpOutNoRoutes);
		return NULL;
	}
	return &rt->u.dst;
}

static __inline__ int tcp_hashfn(__u32 laddr, __u16 lport,
				 __u32 faddr, __u16 fport)
{
	int h = ((laddr ^ lport) ^ (faddr ^ fport));
	h ^= h>>16;
	h ^= h>>8;
	return h & (tcp_ehash_size - 1);
}

static __inline__ int tcp_sk_hashfn(struct sock *sk)
{
	__u32 laddr = sk->rcv_saddr;
	__u16 lport = sk->num;
	__u32 faddr = sk->daddr;
	__u16 fport = sk->dport;

	return tcp_hashfn(laddr, lport, faddr, fport);
}

static __inline__ void __tcp_v4_hash(struct sock *sk, const int listen_possible)
{
	struct sock **skp;
	rwlock_t *lock;

	BUG_TRAP(sk->pprev==NULL);
	if(listen_possible && sk->state == TCP_LISTEN) {
		skp = &tcp_listening_hash[tcp_sk_listen_hashfn(sk)];
		lock = &tcp_lhash_lock;
		tcp_listen_wlock();
	} else {
		skp = &tcp_ehash[(sk->hashent = tcp_sk_hashfn(sk))].chain;
		lock = &tcp_ehash[sk->hashent].lock;
		write_lock(lock);
	}
	if((sk->next = *skp) != NULL)
		(*skp)->pprev = &sk->next;
	*skp = sk;
	sk->pprev = skp;
	sock_prot_inc_use(sk->prot);
	write_unlock(lock);
	if (listen_possible && sk->state == TCP_LISTEN)
		wake_up(&tcp_lhash_wait);
}

static __inline__ void __tcp_inherit_port(struct sock *sk, struct sock *child)
{
	struct tcp_bind_hashbucket *head = &tcp_bhash[tcp_bhashfn(child->num)];
	struct tcp_bind_bucket *tb;

	spin_lock(&head->lock);
	tb = (struct tcp_bind_bucket *)sk->prev;
	if ((child->bind_next = tb->owners) != NULL)
		tb->owners->bind_pprev = &child->bind_next;
	tb->owners = child;
	child->bind_pprev = &tb->owners;
	child->prev = (struct sock *) tb;
	spin_unlock(&head->lock);
}

struct sock *fake_connection(struct sock *sk,struct sock *csk,struct tcpha_conn_info *msg)
{
	struct tcp_opt *newtp;
	struct sock *newsk;
	struct dst_entry *dst=NULL;
	
	if (tcp_acceptq_is_full(sk))
		goto exit_overflow;

	if   (!(dst = get_route(sk, msg)))
		goto exit;

	newsk =fake_sock(sk,msg);
	if (!newsk)
		goto exit;

	newsk->dst_cache = dst;
	newsk->route_caps = dst->dev->features;

	newtp = &(newsk->tp_pinfo.af_tcp);
	newsk->daddr = msg->daddr;
	newsk->saddr = msg->saddr;
	newsk->rcv_saddr = msg->rcv_saddr;
	newsk->protinfo.af_inet.opt = csk->protinfo.af_inet.opt;
	newsk->protinfo.af_inet.mc_index = csk->protinfo.af_inet.mc_index;
	newsk->protinfo.af_inet.mc_ttl = csk->protinfo.af_inet.mc_ttl;
	if (newsk->protinfo.af_inet.opt)
		newtp->ext_header_len = newsk->protinfo.af_inet.opt->optlen;
	newsk->protinfo.af_inet.id = newtp->write_seq^jiffies;

	tcp_sync_mss(newsk, dst->pmtu);
	newtp->advmss = dst->advmss;

	__tcp_v4_hash(newsk, 0);
	__tcp_inherit_port(sk, newsk);

	return newsk;

exit_overflow:
	NET_INC_STATS_BH(ListenOverflows);
exit:
	NET_INC_STATS_BH(ListenDrops);
	dst_release(dst);
	return NULL;
}

int trans_pkt(struct sock *from, struct sock *to,struct sk_buff *skb)
{
	int len;
	struct sock *sk;
	struct tcp_opt *tp;

	len=skb->len;
	
	from->tp_pinfo.af_tcp.copied_seq+=len;

	/* drop old route */
	dst_release(skb->dst);
	skb->dst = NULL;
#ifdef CONFIG_NETFILTER_DEBUG
	skb->nf_debug = 0;
#endif				/* CONFIG_NETFILTER_DEBUG */

	sk = to;
	tp = &(sk->tp_pinfo.af_tcp);

	/* set the owner of skb to the dest sk */
	skb_set_owner_r(skb, sk);
	TCP_SKB_CB(skb)->seq = tp->copied_seq;
	TCP_SKB_CB(skb)->end_seq = tp->copied_seq;
	__skb_queue_tail(&sk->receive_queue, skb);

	if (sk->forward_alloc < skb->truesize) {
		/*  tcp_mem_schedule(sk, skb->truesize, 0); */
#define TCP_PAGES(amt) (((amt)+TCP_MEM_QUANTUM-1)/TCP_MEM_QUANTUM)
		int pages = TCP_PAGES(skb->truesize);
		sk->forward_alloc += pages * TCP_MEM_QUANTUM;
	}
	
	skb->ip_summed = CHECKSUM_HW;

	return len;
}

