/*
  *  
  *	see tcpha_fe.c for infoes about TCPHA
  *
  */

#ifndef _TCPHA_FE_H
#define _TCPHA_FE_H

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/list.h>                 /* for struct list_head */
#include <linux/spinlock.h>             /* for struct rwlock_t */
#include <linux/skbuff.h>               /* for struct sk_buff */
#include <linux/ip.h>                   /* for struct iphdr */
#include <asm/atomic.h>                 /* for struct atomic_t */
#include <linux/sysctl.h>
#include <linux/net.h>
#include <linux/inet.h>
#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/unistd.h>
#include <linux/wait.h>
#include <asm/unistd.h>
#include <linux/list.h>
#include <linux/file.h>
#include <linux/in.h>
#include <linux/random.h>
#include <linux/smp_lock.h>
#include <linux/proc_fs.h>
#include <linux/netdevice.h>		/* for struct neighbour; */
#include <net/dst.h>			/* for struct dst_entry */
#include <net/route.h>			/* for ip_route_output */
#include <net/tcp.h>
#include <net/udp.h>
#include <linux/types.h>
#include <linux/errno.h>          
#include <net/icmp.h>             /* for icmp_send */
#include <net/ip.h>
#include <net/route.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>

#include "regex/regex.h"

/*
 *      TCPHA_FE State Values
 */
enum {
	TCPHA_FE_S_NONE = 0,
	TCPHA_FE_S_ESTABLISHED,
	TCPHA_FE_S_HANDOFFING,
	TCPHA_FE_S_HANDOFFED,
	TCPHA_FE_S_TIME_WAIT,
	TCPHA_FE_S_CLOSE,
	TCPHA_FE_S_CLOSE_WAIT,
	TCPHA_FE_S_ICMP,
	TCPHA_FE_S_LAST
};

enum {
	TCPHA_BE_S_AVAIL =0,
	TCPHA_BE_S_NOTAVAIL,
	TCPHA_BE_S_BUSY
};

/* infoes used by TCP connection Handoff */
struct tcpha_conn_info {
	__u32 			magic;		/* Magic number */
	__u32			connmagic;	/* Connection magic number */
	__u32			daddr;		/* Foreign IPv4 addr			*/
	__u32 			saddr;		/* Sending source			*/
	__u32			rcv_saddr;	/* Bound local IPv4 addr		*/
	__u16			dport;		/* Destination port			*/
	__u16			sport;		/* Source port				*/
	struct tcp_opt	af_tcp;
};

#define TCPHACONNINFOSIZE sizeof(struct tcpha_conn_info)

struct tcpha_fe_info {
	__u32 			magic;
	__u32 			caddr;
	__u16			cport;
	__u32			daddr;
	__u16			dport;
	__u32			ndaddr;
};

struct tcpha_fe_config {
	/* the server IP address and/or port */
	__u32 raddr;
	__u32 vaddr;
	__u16 port;
	__u16 advport;
	
	int startservers;
	int maxSpareServers;
	int minSpareServers;

	int maxClients;
	int conn_perbe;

	/* address/port to redirect */
	__u32 redirect_addr;
	__u16 redirect_port;
};

typedef struct {
	struct list_head list;
	struct tcpha_be *be;
}be_t;

typedef struct {
	struct list_head list;
	regex_t rx;
	struct list_head be_list;
	spinlock_t __be_list_lock;
}rule_t;

/*
 *	Transport protocol header
 */
union tcpha_fe_tphdr {
	unsigned char *raw;
	struct tcphdr *th;
	struct icmphdr *icmph;
	__u16 *portp;
};

struct tcpha_fe_service {

	/* server configuration */
	struct tcpha_fe_config conf;
	
	/* 	real server list and rule list,
	  *	no need locks since once they are  initilized,
	  *	won't change later 
	  */
	struct list_head be_list;
	struct list_head rule_list;
	/* server control */
	int start;
	int stop;

	/* run-time variables */
	struct socket *mainsock;
	struct socket *advsock;
	atomic_t workercount;	/* worker counter */
	atomic_t running;	/* running flag */
};

struct tcpha_fe_worker {
	struct tcpha_fe_service *svc;	/* service it belongs to */
	int pid;		/* pid of worker */
	volatile __u16 status;	/* worker status */
};

struct timeout_table {
	atomic_t refcnt;
	int scale;
	int timeout[TCPHA_FE_S_LAST+1];
};

/* a connection to client */
struct tcpha_fe_conn {
	struct list_head list;	/* d-linked list head */
	rwlock_t lock;
	
	__u32 	caddr;		/* client address */
	__u16  	cport;		/* client port */
	__u32     vaddr;		/* FE virtual address */
	__u32     raddr;		/* FE real address */
	__u16     vport;		/* FE port */
	__u32 	daddr;		/* BE address */
	__u16    	dport;		/* BE port */

	atomic_t		refcnt;		/* reference count */
	struct timer_list	timer;		/* Expiration timer */
	volatile unsigned long	timeout;	/* timeout */
	struct timeout_table *timeout_table;

	volatile __u16          flags;          /* status flags */
	volatile __u16          state;          /* state info */
	
	struct socket *csock;	/* socket connected to client */

	char *buffer;		/* buffer for conn handling */
	size_t buflen;		/* buffer length */

	int (*packet_xmit)(struct sk_buff *skb, struct tcpha_fe_conn *cp);

	#ifdef CONFIG_TCPHA_FE_DEBUG
	int magic;
	#endif
};

/* a connection to BE */
struct tcpha_control_conn {
	struct list_head list;
	struct socket *sock;
	struct tcpha_be *be;

	#ifdef CONFIG_TCPHA_FE_DEBUG
	int magic;					/* for debug */
	#endif
};

/* cluster BE entry */
struct tcpha_be {
	struct list_head list;
	
	struct list_head conn_list; /* keep alive control connection pool */
	atomic_t conn_count;
	spinlock_t __conn_list_lock;

	volatile __u16          state;          /* state info */
	__u32 addr;		/* IP address of real server */
	__u16 port;		/* port number of the service */
};

#define TCPHAMAGIC 		0x12968b9

/*
 *      TCPHA Connection Flags
 */
#define TCPHA_FE_CONN_F_FWD_MASK         0x0007    /* mask for the fwd methods */
#define TCPHA_FE_CONN_F_TUNNEL	      0x0002    /* tunneling */
#define TCPHA_FE_CONN_F_HASHED	      0x0040	/* hashed entry */

/*
 *     TCPHA_FE connection entry hash table
 */
#ifndef CONFIG_TCPHA_FE_TAB_BITS
#define CONFIG_TCPHA_FE_TAB_BITS   12
#endif
/* make sure that TCPHA_FE_CONN_TAB_BITS is located in [8, 20] */
#if CONFIG_TCPHA_FE_TAB_BITS < 8
#define TCPHA_FE_CONN_TAB_BITS	8
#endif
#if CONFIG_TCPHA_FE_TAB_BITS > 20
#define TCPHA_FE_CONN_TAB_BITS	20
#endif
#if 8 <= CONFIG_TCPHA_FE_TAB_BITS && CONFIG_TCPHA_FE_TAB_BITS <= 20
#define TCPHA_FE_CONN_TAB_BITS	CONFIG_TCPHA_FE_TAB_BITS
#endif
#define TCPHA_FE_CONN_TAB_SIZE     (1 << TCPHA_FE_CONN_TAB_BITS)
#define TCPHA_FE_CONN_TAB_MASK     (TCPHA_FE_CONN_TAB_SIZE - 1)

#define INFO(msg...) printk(KERN_INFO "TCPHA_FE: " msg)
#ifdef CONFIG_TCPHA_FE_DEBUG
#define DBG(msg...) printk(KERN_INFO "TCPHA_FE: " msg)
#define DBG_RL(msg...) 					\
    do {						\
	    if (net_ratelimit())			\
		    printk(KERN_INFO "TCPHA_FE: " msg);	\
    } while (0)
#else
#define DBG(msg...) do {} while (0)
#define DBG_RL(msg...) do {} while (0)
#endif
#define ERR(msg...) printk(KERN_ERR "TCPHA_FE: " msg)
#define ERR_RL(msg...)				\
    do {						\
	    if (net_ratelimit())			\
		    printk(KERN_ERR "TCPHA_FE: " msg);	\
    } while (0)

/*
 *  Connection hash table: for input and output packets lookups of TCPHA_FE
 */
struct list_head *tcpha_fe_conn_tab;

/* SLAB cache for TCPHA_FE connections */
kmem_cache_t *tcpha_fe_conn_cachep;

/* counter for current TCPHA_FE connections */
atomic_t tcpha_fe_conn_count;

/* random value for TCPHA_FE connection hash */
unsigned int tcpha_fe_conn_rnd;

struct tcpha_fe_service tfs;

#endif
