/*
  *  
  *	see tcpha_fe.c for infoes about TCPHA
  *
  */
  
#include "tcpha_fe.h"

int tcphafe_max_backlog = 2048;

int
SetupServerSocket(struct tcpha_fe_service *svc)
{
	struct socket *sock;
	struct sockaddr_in sin;
	int error;

	/* First create a socket */
	error = sock_create(PF_INET, SOCK_STREAM, IPPROTO_TCP, &sock);
	if (error < 0) {
		ERR("Error during creation of socket; terminating\n");
		return error;
	}

	/* Now bind the socket */
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_port = svc->conf.port;

	/* set the option to reuse the address. */
	sock->sk->reuse = 1;

	error =
	    sock->ops->bind(sock, (struct sockaddr *) &sin, sizeof(sin));
	if (error < 0) {
		ERR("Error binding socket. This means that some "
			   "other daemon is (or was a short time ago) "
			   "using %u.%u.%u.%u:%d.\n",
			   NIPQUAD(svc->conf.vaddr), ntohs(svc->conf.port));
		return error;
	}

	/* Now, start listening on the socket */
	error = sock->ops->listen(sock, tcphafe_max_backlog);
	if (error != 0) {
		ERR("Error listening on socket \n");
		return error;
	}

	svc->mainsock = sock;

	/* First create a socket */
	error = sock_create(PF_INET, SOCK_STREAM, IPPROTO_TCP, &sock);
	if (error < 0) {
		ERR("Error during creation of socket; terminating\n");
		return error;
	}

	/* Now bind the socket */
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = svc->conf.raddr;
	sin.sin_port = svc->conf.advport;

	/* set the option to reuse the address. */
	sock->sk->reuse = 1;

	error =
	    sock->ops->bind(sock, (struct sockaddr *) &sin, sizeof(sin));
	if (error < 0) {
		ERR("Error binding socket. This means that some "
			   "other daemon is (or was a short time ago) "
			   "using %u.%u.%u.%u:%d.\n",
			   NIPQUAD(svc->conf.raddr), ntohs(svc->conf.advport));
		return error;
	}

	/* Now, start listening on the socket */
	error = sock->ops->listen(sock, tcphafe_max_backlog);
	if (error != 0) {
		ERR("Error listening on socket \n");
		return error;
	}

	svc->advsock = sock;
	
	return 0;
}

void
ReleaseServerSocket(struct tcpha_fe_service *svc)
{
	struct socket *sock;

	if (svc->mainsock) {
		sock = svc->mainsock;
		svc->mainsock = NULL;
		sock_release(sock);
	}
	if (svc->advsock) {
		sock=svc->advsock;
		svc->advsock=NULL;
		sock_release(sock);
	}
}

struct socket *
tcpha_fe_connect2be(struct tcpha_be * be)
{

	struct socket *sock;
	struct sockaddr_in sin;
	int error;

	/* First create a socket */
	error = sock_create(PF_INET, SOCK_STREAM, IPPROTO_TCP, &sock);
	if (error < 0) {
		ERR("Error during creation of socket; terminating\n");
		return NULL;
	}

	/* Now connect to the destination server */
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = be->addr;
	sin.sin_port = be->port;

	error = sock->ops->connect(sock, (struct sockaddr *) &sin,
				   sizeof(sin), 0);
	if (error < 0) {
		ERR("Error connecting to the remote host\n");
		sock_release(sock);
		return NULL;
	}

	return sock;
}

/*
 * tcpha_fe_sendbuffer is to send bytes from the buffer to the socket.
 *
 * A positive return-value indicates the number of bytes sent, a negative
 * value indicates an error-condition.
 */
int
tcpha_fe_sendbuffer(struct socket *sock, const char *buffer,
		  const size_t length, unsigned long flags)
{
	struct msghdr msg;
	mm_segment_t oldfs;
	struct iovec iov;
	int len;

	msg.msg_name = 0;
	msg.msg_namelen = 0;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags = MSG_NOSIGNAL | flags;
	msg.msg_iov->iov_base = (void *) buffer;
	msg.msg_iov->iov_len = length;

	oldfs = get_fs();
	set_fs(KERNEL_DS);
	len = sock_sendmsg(sock, &msg, length);
	set_fs(oldfs);

	return len;
}

int
tcpha_fe_recvbuffer(struct socket *sock, char *buffer,
		  const size_t buflen, unsigned long flags)
{
	struct msghdr msg;
	struct iovec iov;
	int len;
	mm_segment_t oldfs;

	/* Receive a packet */
	msg.msg_name = 0;
	msg.msg_namelen = 0;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags = 0;

	iov.iov_base = buffer;
	iov.iov_len = (size_t) buflen;

	oldfs = get_fs();
	set_fs(KERNEL_DS);
	len = sock_recvmsg(sock, &msg, buflen, flags);
	set_fs(oldfs);

	if (len < 0)
		return -1;

	return len;
}
