/*
  *  
  *	see tcpha_fe.c for infoes about TCPHA
  *
  */
  
#include "tcpha_fe.h"

#define TCPHA_FE_FWD_METHOD(cp)  (cp->flags & TCPHA_FE_CONN_F_FWD_MASK)

static int tcpha_fe_null_xmit(struct sk_buff *skb, struct tcpha_fe_conn *cp)
{
	return NF_ACCEPT;
}

static inline struct rtable *tcpha_fe_get_out_rt(__u32 dst, u32 rtos)
{
	struct rtable *rt;			/* Route to the other host */

       if (ip_route_output(&rt, dst, 0, rtos, 0)) {
		return NULL;
	}

	return rt;
}

static int tcpha_fe_tunnel_xmit(struct sk_buff *skb, struct tcpha_fe_conn *cp)
{
	struct rtable *rt;			/* Route to the other host */
	struct net_device *tdev;		/* Device to other host */
	struct iphdr  *old_iph = skb->nh.iph;
	u8     tos = old_iph->tos;
	u16    df = old_iph->frag_off;
	struct iphdr  *iph;			/* Our new IP header */
	int    max_headroom;			/* The extra header space needed */
	int    mtu;

	if (skb->protocol != __constant_htons(ETH_P_IP)) {
		DBG_RL("tcpha_fe_tunnel_xmit(): protocol error, "
			     "ETH_P_IP: %d, skb protocol: %d\n",
			     __constant_htons(ETH_P_IP), skb->protocol);
		goto tx_error;
	}

	if (!(rt = tcpha_fe_get_out_rt(cp->daddr, RT_TOS(tos))))
		goto tx_error_icmp;

	tdev = rt->u.dst.dev;

	mtu = rt->u.dst.pmtu - sizeof(struct iphdr);
	if (mtu < 68) {
		ip_rt_put(rt);
		DBG_RL("tcpha_fe_tunnel_xmit(): mtu less than 68\n");
		goto tx_error;
	}
	if (skb->dst && mtu < skb->dst->pmtu)
		skb->dst->pmtu = mtu;

	df |= (old_iph->frag_off&__constant_htons(IP_DF));

	if ((old_iph->frag_off&__constant_htons(IP_DF))
	    && mtu < ntohs(old_iph->tot_len)) {
		icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
		ip_rt_put(rt);
		DBG_RL("tcpha_fe_tunnel_xmit(): frag needed\n");
		goto tx_error;
	}

	/* update checksum because skb might be defragmented */
	ip_send_check(old_iph);

	skb->h.raw = skb->nh.raw;

	/*
	 * Okay, now see if we can stuff it in the buffer as-is.
	 */
	max_headroom = (((tdev->hard_header_len+15)&~15)+sizeof(struct iphdr));

	if (skb_headroom(skb) < max_headroom
	    || skb_cloned(skb) || skb_shared(skb)) {
		struct sk_buff *new_skb =
			skb_realloc_headroom(skb, max_headroom);
		if (!new_skb) {
			ip_rt_put(rt);
			kfree_skb(skb);
			ERR_RL("tcpha_fe_tunnel_xmit(): no memory\n");
			return -EINVAL;
		}
		kfree_skb(skb);
		skb = new_skb;
		old_iph = skb->nh.iph;
	}

	skb->nh.raw = skb_push(skb, sizeof(struct iphdr));
	memset(&(IPCB(skb)->opt), 0, sizeof(IPCB(skb)->opt));

	/* drop old route */
	dst_release(skb->dst);
	skb->dst = &rt->u.dst;

	/*
	 *	Push down and install the IPIP header.
	 */
	iph			=	skb->nh.iph;
	iph->version		=	4;
	iph->ihl		=	sizeof(struct iphdr)>>2;
	iph->frag_off		=	df;
	iph->protocol		=	IPPROTO_IPIP;
	iph->tos		=	tos;
	iph->daddr		=	rt->rt_dst;
	iph->saddr		=	cp->raddr;
	iph->ttl		=	old_iph->ttl;
	iph->tot_len		=	htons(skb->len);
	ip_select_ident(iph, &rt->u.dst, NULL);
	ip_send_check(iph);

	skb->ip_summed = CHECKSUM_NONE;
#ifdef CONFIG_NETFILTER_DEBUG
	skb->nf_debug = 1 << NF_IP_LOCAL_OUT;
#endif /* CONFIG_NETFILTER_DEBUG */
	ip_send(skb);

	return NF_STOLEN;

  tx_error_icmp:
	dst_link_failure(skb);
  tx_error:
	kfree_skb(skb);
	return NF_STOLEN;
}

void tcpha_fe_bind_xmit(struct tcpha_fe_conn *cp)
{
	switch (TCPHA_FE_FWD_METHOD(cp)) {
	case TCPHA_FE_CONN_F_TUNNEL:
		cp->packet_xmit = tcpha_fe_tunnel_xmit;
		break;
	default:
		cp->packet_xmit = tcpha_fe_null_xmit;
		break;
	}
}
